String.prototype.isEmpty = function () {
    return (!this || /^\s*$/.test(this));
}

// extends the dialog widget with a new option: adding new buttons to the title bar
$.widget('app.dialog', $.ui.dialog, {
    options: {
        iconButtons: []
    },
    _create: function () {
        // call the default widget constructor
        this._super();

        var index = 0;
        var $titlebar = this.uiDialog.find('.ui-dialog-titlebar');
        $.each(this.options.iconButtons, function () {
         	if (this.show) {
	            // finds the last button added; this is actually the left-most button
	            var $button = $('<button/>').text(this.text);
	            var $closeButton = $titlebar.find('[type="button"]:last');
	            var right = $closeButton.css('right');
	            var closeButtonWidth = $closeButton.outerWidth(true);
	            var rightCss = (parseFloat(right) + 1.1 * (index++ + 1) * closeButtonWidth) + 'px';
	
	            // creates the new button widget, adding it to the title bar
	            $button.button({
	            	disabled: this.disabled,
	            	icons: {
	            		primary: this.icon
	            	},
	            	text: false
	        	}).addClass('ui-dialog-titlebar-close')
	        		.addClass(this.clazz)
	        		.css('right', rightCss)
	        		.click(this.click)
	        		.appendTo($titlebar);
        	}
        });
    }
});

$.throttle = (func, limit) => {
	let lastFunc;
	let lastRan;
	return function () {
		const context = this;
	    const args = arguments;
	    if (!lastRan) {
	    	func.apply(context, args);
	    	lastRan = Date.now();
	    } else {
	    	clearTimeout(lastFunc);
	    	lastFunc = setTimeout(function() {
		        if ((Date.now() - lastRan) >= limit) {
		        	func.apply(context, args);
		        	lastRan = Date.now();
		        }
	    	}, limit - (Date.now() - lastRan))
	    }
	}
};

$.debounce = (func, delay) => {
	let inDebounce;
	return function () {
		const context = this;
		const args = arguments;
		clearTimeout(inDebounce);
		inDebounce = setTimeout(() => func.apply(context, args), delay);
  	}
}

$.clazz = function (clazz) {
	return ('.' + clazz);
}

$.id = function (id) {
	return ('#' + id.replace(/\[|\]|\s/g, ''));
}

function calculateDxy(id) {
	var $div = $($.id(id));
	var border = $div.parent().outerWidth() - $div.parent().innerWidth();
	var padding = $div.parent().innerWidth() - $div.parent().width();
	var margin = $div.parent().outerWidth(true) - $div.parent().outerWidth();
	var dialogTitleHeight = $div.parent().children().first().outerHeight(true) || 0;
	var dx = (border + padding + margin)/2;
	var dy = (border + padding + margin)/2;
	return [dx, dy, dialogTitleHeight];
}

function drawCanvasRoundedRectangle(context, x, y, width, height, radius, lineWidth, fillStyle, strokeStyle) {
	context.beginPath();
	context.lineWidth = lineWidth;
	context.strokeStyle = strokeStyle;
	context.fillStyle = fillStyle;
	context.moveTo(x + radius, y);
	context.lineTo(x + width - radius, y);
	context.quadraticCurveTo(x + width, y, x + width, y + radius);
	context.lineTo(x + width, y + height - radius);
	context.quadraticCurveTo(x + width, y + height, x + width - radius, y + height);
	context.lineTo(x + radius, y + height);
	context.quadraticCurveTo(x, y + height, x, y + height - radius);
	context.lineTo(x, y + radius);
	context.quadraticCurveTo(x, y, x + radius, y);
	context.closePath();
	context.fill();
	context.stroke();
}

function fullScreenProcessor(fullScreen, browserFullScreenEvent) {
	var $applicationsBar = $($.clazz('applications-bar'));
	var $barsLayout = $($.clazz('bars-layout'));
	var $navigationBar = $($.clazz('navigation-bar'));
	var $poweredByDiv = $($.clazz('powered-by-dvirtual'));
	var $runningApplicationsBar = $($.clazz('running-applications-bar'));
	var $windowsBar = $($.clazz('windows-bar'));

	var $fullScreenBottomApplications = $($.clazz(FULL_SCREEN_BOTTOM_APPLICATIONS_CLASS));
	var $fullScreenBottomBars = $($.clazz(FULL_SCREEN_BOTTOM_CLASS));
	var $fullScreenElement = $($.clazz(FULL_SCREEN_ELEMENT_CLASS));
	var $fullScreenTopApplications = $($.clazz(FULL_SCREEN_TOP_APPLICATIONS_CLASS));
	var $fullScreenTopBars = $($.clazz(FULL_SCREEN_TOP_CLASS));
	
	var animationDelay = 500;
	var applicationsTitle = 'Applications';
	var runningApplicationsBarOverflow = $runningApplicationsBar.find('vaadin-tabs').attr('overflow');
	var windowsBarOverflow = $windowsBar.find('vaadin-tabs').attr('overflow');
	
	if (FULL_SCREEN) {
		if ($fullScreenBottomBars.length === 0 || $fullScreenTopBars.length === 0) {
			// hide bars
			$barsLayout.hide();
			$navigationBar.hide();
			$applicationsBar.hide();

			// set title of $fullScreenTopApplications element
			var title = $('vaadin-button[class*="active"]').attr('applications');
			applicationsTitle = title !== undefined ? title : TRANSLATION_APPLICATIONS;

			// add top application div element
			var top = $applicationsBar.outerHeight(true) + $navigationBar.outerHeight(true);
			$fullScreenTopBars = $('<div />').hide().addClass(FULL_SCREEN_TOP_CLASS).css('top', -top).click(function (event) {
				if (!$(event.target).is('vaadin-tabs')) {
					$fullScreenTopApplications.find('span').delay(animationDelay).fadeOut(10);
					$fullScreenTopApplications.animate({top: '-5px'}, {
						queue: false,
						done: function () {
							var title = $('vaadin-button[class*="active"]').attr('applications');
							applicationsTitle = title !== undefined ? title : TRANSLATION_APPLICATIONS;
							$fullScreenTopApplications.children('div').first().text(applicationsTitle);
					}});
					$fullScreenTopBars.delay(animationDelay).animate({top: '-100px' });
				}
			}).append($navigationBar).append($applicationsBar);
			$fullScreenTopApplications = $('<div />').addClass(FULL_SCREEN_TOP_APPLICATIONS_CLASS)
				.append($('<div />').text(applicationsTitle))
				.append($('<span />').hide().addClass('fullscreen-top-close').html('<iron-icon icon="vaadin:close-circle-o"></iron-icon>'));
			$fullScreenTopApplications.mouseenter(function () {
				if ($fullScreenTopBars.css('top').startsWith('-') === true) {
					// top applications bar is not visible
					$fullScreenTopApplications.animate({top: '-1px'}, {queue: false});
				}
			}).mouseleave(function () {
				if ($fullScreenTopBars.css('top').startsWith('-') === true) {
					// top applications bar is not visible
					$fullScreenTopApplications.animate({top: '-5px'}, {queue: false});
				}
			}).click(function () {
				disableBars(true);
				var title = $('vaadin-button[class*="active"]').attr('applications');
				applicationsTitle = title !== undefined ? title : TRANSLATION_APPLICATIONS;
				if ($fullScreenTopBars.css('top').startsWith('-') === true) {
					// top applications bar is not visible
					$fullScreenTopApplications.children('div').first().text(applicationsTitle);
					$fullScreenTopApplications.find('span').delay(animationDelay).fadeIn(10);
					$fullScreenTopApplications.delay(animationDelay).animate({top: $fullScreenTopBars.outerHeight(true) - 5});
					$fullScreenTopBars.animate({top: '0px'}, {queue: false});
				} else {
					$fullScreenTopApplications.find('span').delay(animationDelay).fadeOut(10);
					$fullScreenTopApplications.animate({top: '-5px'}, {queue: false}, function () {
						$fullScreenTopApplications.find('span').hide();
					});
					$fullScreenTopBars.delay(animationDelay).animate({top: '-100px' });
					$fullScreenTopApplications.children('div').first().text(applicationsTitle);
				}
				setTimeout(function() {
					disableBars(false);
				}, 2 * animationDelay);
			}).prependTo($fullScreenElement);
			$fullScreenTopBars.prependTo($fullScreenElement).show();
			
			// add bottom application div element
			var bottom = $barsLayout.outerHeight(true);
			var poweredByHeight = $poweredByDiv.outerHeight();
			$fullScreenBottomBars = $('<div />').hide().addClass(FULL_SCREEN_BOTTOM_CLASS).css('bottom', -bottom).click(function (event) {
				if (!$(event.target).is('vaadin-tabs')) {
					$fullScreenBottomApplications.find('span').delay(animationDelay).fadeOut(10);
					$fullScreenBottomApplications.animate({bottom: poweredByHeight - 5}, {queue: false});
					$fullScreenBottomBars.delay(animationDelay).animate({bottom: poweredByHeight - 100});
				}
			}).append($barsLayout);
			var runningApplicationsTitle = TRANSLATION_RUNNING_APPLICATIONS;
			$fullScreenBottomApplications = $('<div />').addClass(FULL_SCREEN_BOTTOM_APPLICATIONS_CLASS).append($('<div />').text(runningApplicationsTitle))
				.append($('<span />').hide().addClass('fullscreen-bottom-close').html('<iron-icon icon="vaadin:close-circle-o"></iron-icon>'));
			$fullScreenBottomApplications.mouseenter(function () {
				if ($fullScreenBottomBars.css('bottom').startsWith('-') === true) {
					// bottom applications bar is not visible
					$fullScreenBottomApplications.animate({bottom: poweredByHeight - 1}, {queue: false});
				}
			}).mouseleave(function () {
				if ($fullScreenBottomBars.css('bottom').startsWith('-') === true) {
					// bottom applications bar is not visible
					$fullScreenBottomApplications.animate({bottom: poweredByHeight - 5}, {queue: false});
				}
			}).click(function () {
				disableBars(true);
				if ($fullScreenBottomBars.css('bottom').startsWith('-') === true) {
					// bottom applications bar is not visible
					$fullScreenBottomApplications.find('span').delay(animationDelay).fadeIn(10);
					$fullScreenBottomApplications.delay(animationDelay).animate({bottom: $fullScreenBottomBars.outerHeight(true) + poweredByHeight - 5});
					$fullScreenBottomBars.animate({bottom: poweredByHeight}, {queue: false});
				} else {
					$fullScreenBottomApplications.find('span').delay(animationDelay).fadeOut(10);
					$fullScreenBottomApplications.animate({bottom: poweredByHeight - 5}, {queue: false}, function() {
						$fullScreenBottomApplications.find('span').hide();
					});
					$fullScreenBottomBars.delay(animationDelay).animate({bottom: poweredByHeight - 100 });
				}
				setTimeout(function() {
					disableBars(false);
				}, 2 * animationDelay);
			}).prependTo($fullScreenElement);
			$fullScreenBottomBars.appendTo($fullScreenElement).show();
			$fullScreenBottomBars.on('mouseenter', 'iron-icon[class="title"]', function () {
				setTimeout(function run() {
					$('body').find('div[class~="ui-tooltip"]').css('z-index', '999999').appendTo($fullScreenElement);
				}, animationDelay);
			});
			$poweredByDiv.appendTo($fullScreenElement);

			// show top and bottom application div elements
			$fullScreenTopApplications.delay(animationDelay).animate({top: '-5px'});
			$fullScreenBottomApplications.delay(animationDelay).animate({bottom: poweredByHeight - 5});

			// show bars
			$barsLayout.show();
			$navigationBar.show();
			$applicationsBar.show();
		}
		
		if (browserFullScreenEvent) {
			$fullScreenElement.find('[class~="settings"]').addClass('no-border');
			$fullScreenElement.addClass('no-border');
		} else {
			$fullScreenElement.fullscreen();
		}
	} else {
		// hide bars
		$barsLayout.hide();
		$navigationBar.hide();
		$applicationsBar.hide();

		// remove top and bottom applications div element
		var $dvirtualLayout = $($.clazz('dvirtual-layout'));
		$dvirtualLayout.find('[class~="settings"]').removeClass('no-border');
		$applicationsBar.prependTo($dvirtualLayout);
		$barsLayout.appendTo($dvirtualLayout);
		$navigationBar.prependTo($dvirtualLayout);
		$poweredByDiv.appendTo($dvirtualLayout);
		if (browserFullScreenEvent) {
			if (!($fullScreenElement.find($($.clazz('canvas-div'))).is(':hidden') && $fullScreenElement.find('div[class*="fullscreen"]').length > 0)) {
				$fullScreenElement.removeClass('no-border');
			}
		} else {
			$.fullscreen.exit();
		}
		$fullScreenBottomBars.off('mouseenter mouseleave').remove();
		$fullScreenBottomApplications.off('click mouseenter mouseleave').remove();
		$fullScreenTopBars.off('mouseenter mouseleave').remove();
		$fullScreenTopApplications.off('click mouseenter mouseleave').remove();

		// show bars
		$barsLayout.show();
		$navigationBar.show();
		$applicationsBar.show();
	}

	// overflow button of 'runningApplicationsBar'
	if (typeof runningApplicationsBarOverflow !== typeof undefined && runningApplicationsBarOverflow !== false) {
		$runningApplicationsBar.find('vaadin-tabs').attr('overflow', 'end');
	}

	// overflow button of 'windowsBar'
	if (typeof windowsBarOverflow !== typeof undefined && windowsBarOverflow !== false) {
		$windowsBar.find('vaadin-tabs').attr('overflow', 'end');
	}
}

function getInteger(value) {
	return Math.round(parseFloat(value)) || 0;
}

function getServer($element) {
	var $parent = $($element).parent();
	while ($parent.length > 0) {
		if($parent[0].$server !== undefined) {
			return $parent[0].$server;
		}
		$parent = $parent.parent();
	}
	return undefined;
}

function isAnotherRequestActive() {
	// check if can send new request to the server
	return Vaadin.Flow.clients.ROOT.isActive();
}
